//
// Created by nicol on 2018-03-18.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include "esc.h"
#include "utils/socket.h"
#include "utils/buffer.h"
#include "proto.h"
#include "tcp_package.h"
#include "utils/thread.h"
#include "utils/mutex.h"
#include "utils/string.h"

#ifdef _WIN32
#define sleep Sleep
#endif

typedef int bool_t;
const bool_t BOOL_TRUE = 1;
const bool_t BOOL_FALSE = 0;

#define ERROR_BUFFER_SIZE 4096
esc_error_t error_create(int code, char* format, ...) {
    va_list vl;
    va_start(vl, format);
    char buf[ERROR_BUFFER_SIZE];
    size_t size = (size_t)snprintf(buf, ERROR_BUFFER_SIZE, format, vl);
    char* msg = malloc(size+1);
    strcpy(msg, buf);
    esc_error_t res = {code, msg};
    return res;
}

struct st_connection_settings {
    bool_t use_ssl_connection;
};

const esc_connection_settings_t default_connection_settings = {
};

const esc_connection_settings_t* esc_default_connection_settings = &default_connection_settings;

struct st_tcp_endpoint {
    char* host;
    unsigned short port;
};
typedef struct st_tcp_endpoint tcp_endpoint_t;

struct st_node_endpoints {
    tcp_endpoint_t tcp_endpoint;
    tcp_endpoint_t secure_tcp_endpoint;
};
typedef struct st_node_endpoints node_endpoints_t;

typedef const node_endpoints_t* (*endpoint_discoverer_t)(const void* discover_data, const tcp_endpoint_t* failed_tcp_endpoint);

struct st_connection {
    esc_connection_settings_t settings;
    const char* name;
    void* discoverer_data;
    endpoint_discoverer_t discover;
    socket_t tcp_conn;
    ProtobufCAllocator protobuf_c_allocator;
    thread_t manager_thread;
	mutex_t sending_lock;
    mutex_t send_lock;
    tcp_package_t* send_pkg;
    mutex_t recv_lock;
    tcp_package_t* recv_pkg;
};

struct st_static_endpoint_discoverer {
    tcp_endpoint_t tcp_endpoint;
    bool_t use_ssl_connection;
};
typedef struct st_static_endpoint_discoverer static_endpoint_discoverer_t;

const node_endpoints_t* static_discover(const static_endpoint_discoverer_t* discover_data, const tcp_endpoint_t* failed_tcp_endpoint) {
    node_endpoints_t* result = malloc(sizeof(node_endpoints_t));
    if (discover_data->use_ssl_connection) {
        result->secure_tcp_endpoint = discover_data->tcp_endpoint;
    } else {
        result->tcp_endpoint = discover_data->tcp_endpoint;
    }
    return result;
}

int connection_send_tcp_package(const esc_connection_t* conn, const tcp_package_t* pkg) {
    char uuid_buf[37];
    fprintf(stderr, "connection_send_tcp_package: %u %u %s %lu\n", pkg->command, pkg->flags, esc_uuid_format(&pkg->correlation_id, uuid_buf, 37), pkg->data.size);
    buffer_t send_buffer = tcp_package_to_buffer(pkg);
    uint32_t size = (uint32_t)send_buffer.size;
    if (socket_send(conn->tcp_conn, (char *) &size, sizeof(uint32_t)) <= 0) {
        buffer_free(send_buffer);
        return -1;
    }
    if (socket_send(conn->tcp_conn, (char *) send_buffer.data, send_buffer.size) <= 0) {
        buffer_free(send_buffer);
        return -2;
    }
    buffer_free(send_buffer);
    return 0;
}

const tcp_package_t* connection_recv_tcp_package(const esc_connection_t* conn) {
    uint32_t recv_size;
    ssize_t rc;
    if ((rc = socket_recv(conn->tcp_conn, (char *)&recv_size, sizeof(uint32_t))) != 4) {
        fprintf(stderr, "%ld %d", rc, socket_error());
        return 0;
    }
    buffer_t recv_buffer = buffer_create(recv_size);
    while(recv_size > 0) {
        size_t pos = recv_buffer.size - recv_size;
        rc = socket_recv(conn->tcp_conn, (char *)&recv_buffer.data[pos], recv_size);
        recv_size -= rc;
    }
    const tcp_package_t* recv_pkg = tcp_package_from_buffer(recv_buffer);
    buffer_free(recv_buffer);
    char uuid_buf[37];
    fprintf(stderr, "connection_recv_tcp_package: %u %u %s %lu\n", recv_pkg->command, recv_pkg->flags, esc_uuid_format(&recv_pkg->correlation_id, uuid_buf, 37), recv_pkg->data.size);
    //for (int32_t i=0;i<recv_pkg->data.size;i++) {
    //    printf("%x (%c) ", recv_pkg->data.data[i], recv_pkg->data.data[i]);
    //}
    //printf("\n");
    return recv_pkg;
}

void connection_enqueue_send(esc_connection_t* conn, tcp_package_t* pkg) {
    mutex_lock(conn->sending_lock);	
	mutex_lock(conn->send_lock);
    conn->send_pkg = pkg;
	mutex_unlock(conn->send_lock);
}

tcp_package_t* connection_wait_for(esc_connection_t* conn, esc_uuid_t* correlation_id) {
    tcp_package_t* found = 0;
    while(found == 0) {
        mutex_lock(conn->recv_lock);
        if (conn->recv_pkg && esc_uuid_compare(&conn->recv_pkg->correlation_id, correlation_id) == 0) {
            found = conn->recv_pkg;
			conn->recv_pkg = 0;
        }
        mutex_unlock(conn->recv_lock);
		sleep(10);
    }
    return found;
}

void* protobuf_c_alloc(void *alloc_data, size_t size) {
    return malloc(size);
}

void protobuf_c_free(void *alloc_data, void* p) {
    free(p);
}

const esc_connection_t* esc_connection_create(const esc_connection_settings_t* connection_settings, const char* addr, const char* connection_name) {
    if (strncmp(addr, "tcp://", 6) != 0) {
        return 0;
    }
    char* pos = strrchr(addr, ':');
    if (pos == 0) {
        return 0;
    }
    char* host = malloc(pos - addr - 5);
    strncpy(host, addr+6, pos-addr-6);
    host[pos-addr-6] = 0;
    unsigned short port = (unsigned short)atoi(pos+1);
    static_endpoint_discoverer_t discover_data = {
        {host, port}, connection_settings->use_ssl_connection
    };
    struct st_connection* conn = malloc(sizeof(struct st_connection));
    conn->settings = *connection_settings;
    conn->discoverer_data = malloc(sizeof(static_endpoint_discoverer_t));
    memcpy(conn->discoverer_data, &discover_data, sizeof(static_endpoint_discoverer_t));
    conn->discover = (endpoint_discoverer_t)static_discover;
    conn->protobuf_c_allocator.alloc = protobuf_c_alloc;
    conn->protobuf_c_allocator.free = protobuf_c_free;
    conn->protobuf_c_allocator.allocator_data = 0;
    if (connection_name == 0 || strcmp(connection_name, "") == 0) {
        const esc_uuid_t* uuid = esc_uuid_create();
        char buf[40];
        esc_uuid_format(uuid, buf, 40);
        conn->name = string_copy(buf);
        esc_uuid_free(uuid);
    } else {
        conn->name = connection_name;
    }
	conn->recv_pkg = 0;
	conn->send_pkg = 0;
	conn->send_lock = mutex_create();
	conn->sending_lock = mutex_create();
	conn->recv_lock = mutex_create();
    return conn;
}

void* connection_thread(void* arg) {
    esc_connection_t* conn = arg;

    while(1) {
        mutex_lock(conn->send_lock);
		//printf("%p ", conn->send_pkg);		
		int rc;
        if (conn->send_pkg) {
			if ((rc = connection_send_tcp_package(conn, conn->send_pkg)) == 0) {
				// free send pkg
				conn->send_pkg = 0;
				mutex_unlock(conn->sending_lock);
			}
        }
		mutex_unlock(conn->send_lock);
		
		if ((rc = socket_readable(conn->tcp_conn)) > 0) {
			mutex_lock(conn->recv_lock);
			if (conn->recv_pkg == 0) {
				conn->recv_pkg = connection_recv_tcp_package(conn);
			}
			mutex_unlock(conn->recv_lock);
		}
		//printf("%d ", rc);
		
		mutex_lock(conn->recv_lock);
		if (conn->recv_pkg && conn->recv_pkg->command == 0x01) {
			tcp_package_t* hb = tcp_package_create(0x02, &conn->recv_pkg->correlation_id, buffer_create(0));
			conn->recv_pkg = 0;
			connection_send_tcp_package(conn, hb);
		}
		mutex_unlock(conn->recv_lock);

        //mutex_unlock(conn->mutex_lock);
        sleep(10);
    }
}

// return 0 on success
// return non-zero on failure and sets last_error on connection
int esc_connection_connect(const esc_connection_t* conn) {
    // Discover endpoint
    const node_endpoints_t* endpoints = conn->discover(conn->discoverer_data, 0);
    if (endpoints == 0) {
        return -1;
    }
    // Establish Tcp Connection
    esc_connection_t* _conn = (esc_connection_t*)conn;
    _conn->tcp_conn = socket_create(SOCKET_TYPE_TCP);
    tcp_endpoint_t endpoint = conn->settings.use_ssl_connection ? endpoints->secure_tcp_endpoint : endpoints->tcp_endpoint;
    if (socket_connect(conn->tcp_conn, endpoint.host, endpoint.port) != 0) {
        return -2;
    }
	// Start manager thread
    _conn->manager_thread = thread_create(connection_thread, _conn);
    if (thread_start(_conn->manager_thread)) {
        return -3;
    }
	// Identify
    // build message
    EventStore__Client__Messages__IdentifyClient identify_client;
    event_store__client__messages__identify_client__init(&identify_client);
    identify_client.connection_name = (char*)conn->name;
    identify_client.version = 1;
    size_t s = event_store__client__messages__identify_client__get_packed_size(&identify_client);
    uint8_t buffer[s];
    event_store__client__messages__identify_client__pack(&identify_client, buffer);
    // build tcp_package
    const tcp_package_t* send_pkg = tcp_package_create(0xF5, esc_uuid_create(), buffer_from(buffer, s));
    connection_enqueue_send(_conn, send_pkg);
	const tcp_package_t* recv_pkg = connection_wait_for(_conn, &send_pkg->correlation_id);
    if (recv_pkg->command != 0xF6) {
        return -5;
    }
	
	return 0;
}

const esc_credentials_t* esc_credentials_create(const char* username, const char* password) {
    esc_credentials_t* creds = malloc(sizeof(struct st_credentials));
    creds->username = username;
    creds->password = password;
    return creds;
}

const esc_recorded_event_t* recorded_event_create(EventStore__Client__Messages__EventRecord* msg) {
    if (msg == 0) return 0;
    esc_recorded_event_t* ev = malloc(sizeof(esc_recorded_event_t));
    ev->event_id = esc_uuid_from(msg->event_id.data, msg->event_id.len);
    ev->event_type = string_copy(msg->event_type);
    ev->event_number = msg->event_number;
    ev->event_stream_id = string_copy(msg->event_stream_id);
    ev->created_epoch = msg->has_created_epoch ? msg->created_epoch : 0;
    ev->data = buffer_from(msg->data.data, msg->data.len);
    ev->metadata = buffer_from(msg->metadata.data, msg->metadata.len);

    return ev;
}

esc_resolved_event_t* resolved_event_create(EventStore__Client__Messages__ResolvedEvent* msg) {
    esc_resolved_event_t* ev = malloc(sizeof(esc_resolved_event_t));
    ev->original_position.prepare_position = msg->prepare_position;
    ev->original_position.commit_position = msg->commit_position;
    ev->event = recorded_event_create(msg->event);
    ev->link = recorded_event_create(msg->link);
    return ev;
}

const esc_all_events_slice_t* esc_connection_read_all_forward(const esc_connection_t* conn, const esc_position_t* last_checkpoint, unsigned int count, const esc_credentials_t* credentials) {
    EventStore__Client__Messages__ReadAllEvents send_msg;
    event_store__client__messages__read_all_events__init(&send_msg);
    send_msg.prepare_position = last_checkpoint ? last_checkpoint->prepare_position : 0;
    send_msg.commit_position = last_checkpoint ? last_checkpoint->commit_position : 0;
    send_msg.max_count = count;
    send_msg.require_master = 0;
    send_msg.resolve_link_tos = 0;
    size_t s = event_store__client__messages__read_all_events__get_packed_size(&send_msg);
    uint8_t buffer[s];
    event_store__client__messages__read_all_events__pack(&send_msg, buffer);
    const tcp_package_t* send_pkg = tcp_package_create_authenticated(0xB6, esc_uuid_create(), buffer_from(buffer, s), credentials->username, credentials->password);
    connection_enqueue_send(conn, send_pkg);

    const tcp_package_t* recv_pkg = connection_wait_for(conn, &send_pkg->correlation_id);
    if (recv_pkg->command != 0xB7) {
        return 0;
    }

    esc_connection_t* _conn = (esc_connection_t*)conn;

    EventStore__Client__Messages__ReadAllEventsCompleted *recv_msg =
        event_store__client__messages__read_all_events_completed__unpack(&_conn->protobuf_c_allocator, recv_pkg->data.size, recv_pkg->data.data);

    esc_all_events_slice_t* result = malloc(sizeof(esc_all_events_slice_t));
    result->read_direction = "forward";
    result->is_end_of_stream = (recv_msg->events == 0 || recv_msg->n_events == 0) ? 1 : 0;
    result->from_position.prepare_position = recv_msg->prepare_position;
    result->from_position.commit_position = recv_msg->commit_position;
    result->next_position.prepare_position = recv_msg->next_prepare_position;
    result->next_position.commit_position = recv_msg->next_commit_position;
    result->n_events = recv_msg->n_events;
    result->events = malloc(recv_msg->n_events * sizeof(esc_resolved_event_t*));
    for (size_t i = 0; i < recv_msg->n_events; i++) {
        result->events[i] = resolved_event_create(recv_msg->events[i]);
    }
    event_store__client__messages__read_all_events_completed__free_unpacked(recv_msg, &_conn->protobuf_c_allocator);

    return result;
}

void esc_connection_close(const esc_connection_t* conn) {
    socket_close(conn->tcp_conn);
}

const char* esc_position_format(const esc_position_t* position, char* buffer, size_t buf_size) {
    snprintf(buffer, buf_size, "%lu/%lu", position->prepare_position, position->commit_position);
    return buffer;
}
